-- Chapter 11 Oracle Examples

11.11
-- SQL *Plus command
SET SERVEROUTPUT ON;
-- Anonymous block
DECLARE
 TmpSum     INTEGER;
 TmpProd    INTEGER;
 Idx        INTEGER;
BEGIN
-- Initialize temporary variables
 TmpSum := 0;
 TmpProd := 1;
 -- Use a loop to compute the sum and product
 FOR Idx IN 1 .. 10 LOOP 
    TmpSum := TmpSum + Idx;
    TmpProd := TmpProd * Idx;
 END LOOP;
 -- Display the results
 Dbms_Output.Put_Line('Sum is ' || To_Char(TmpSum));
 Dbms_Output.Put_Line('Product is ' || To_Char(TmpProd));
END;
/

11.12
SET SERVEROUTPUT ON;
DECLARE
 TmpSum     INTEGER;
 TmpProd    INTEGER;
 Idx        INTEGER;
BEGIN
-- Initialize temporary variables
 TmpSum := 0;
 TmpProd := 1;
 -- Use a loop to compute the sum and product
 -- Mod(X,Y) returns the integer remainder of X/Y.
 FOR Idx IN 1 .. 10 LOOP 
    IF Mod(Idx,2) = 0 THEN -- even number
      TmpSum := TmpSum + Idx;
    ELSE
      TmpProd := TmpProd * Idx;
    END IF;
 END LOOP;
 -- Display the results
 Dbms_Output.Put_Line('Sum is ' || To_Char(TmpSum));
 Dbms_Output.Put_Line('Product is ' || To_Char(TmpProd));
END;
/

11.13
CREATE OR REPLACE PROCEDURE pr_InsertRegistration
(aRegNo IN Registration.RegNo%TYPE,
 aStdSSN IN Registration.StdSSN%TYPE,
 aRegStatus IN Registration.RegStatus%TYPE,
 aRegDate IN Registration.RegDate%TYPE,
 aRegTerm IN Registration.RegTerm%TYPE,
 aRegYear IN Registration.RegYear%TYPE) IS
-- Create a new registration
BEGIN

INSERT INTO Registration 
       (RegNo, StdSSN, RegStatus, RegDate, RegTerm, RegYear)
VALUES (aRegNo, aStdSSN, aRegStatus, aRegDate, aRegTerm, aRegYear);

dbms_output.put_line('Added a row to the Registration table');

END;
/

-- Testing code
SET SERVEROUTPUT ON;
-- Number of rows before the procedure execution
SELECT COUNT(*) FROM Registration;

BEGIN
pr_InsertRegistration
 (1275,'901-23-4567','F',To_Date('27-Feb-2006'),'Spring',2006);
END;
/
-- Number of rows after the procedure execution
SELECT COUNT(*) FROM Registration;
-- Delete the inserted row using the ROLLBACK statement
ROLLBACK;

11.14
CREATE OR REPLACE PROCEDURE pr_InsertRegistration
(aRegNo IN Registration.RegNo%TYPE,
 aStdSSN IN Registration.StdSSN%TYPE,
 aRegStatus IN Registration.RegStatus%TYPE,
 aRegDate IN Registration.RegDate%TYPE,
 aRegTerm IN Registration.RegTerm%TYPE,
 aRegYear IN Registration.RegYear%TYPE,
 aResult OUT BOOLEAN ) IS
-- Create a new registration
-- aResult is TRUE if successful, false otherwise.
BEGIN
aResult := TRUE;
INSERT INTO Registration 
       (RegNo, StdSSN, RegStatus, RegDate, RegTerm, RegYear)
VALUES (aRegNo, aStdSSN, aRegStatus, aRegDate, aRegTerm, aRegYear);


EXCEPTION
WHEN OTHERS THEN aResult := FALSE;
END;
/

-- Testing code
SET SERVEROUTPUT ON;
-- Number of rows before the procedure execution
SELECT COUNT(*) FROM Registration;
DECLARE
 -- Output parameter must be declared in the calling block
 Result BOOLEAN;
BEGIN
-- This test should succeed.
-- Procedure assigns value to the output parameter (Result).
pr_InsertRegistration
(1275,'901-23-4567','F',To_Date('27-Feb-2006'),'Spring',2006,Result);
IF Result THEN
    dbms_output.put_line('Added a row to the Registration table');
ELSE
    dbms_output.put_line('Row not added to the Registration table');
END IF;

-- This test should fail because of the duplicate primary key.
pr_InsertRegistration
(1275,'901-23-4567','F',To_Date('27-Feb-2006'),'Spring',2006,Result);
IF Result THEN
    dbms_output.put_line('Added a row to the Registration table');
ELSE
    dbms_output.put_line('Row not added to the Registration table');
END IF;
END;
/

-- Number of rows after the procedure executions
SELECT COUNT(*) FROM Registration;
-- Delete inserted row
ROLLBACK;

11.15
CREATE OR REPLACE FUNCTION fn_RetrieveStdName
(aStdSSN IN Student.StdSSN%type) RETURN VARCHAR2 IS
-- Retrieves the student name (concatenate first and last name)
-- given a student SSN. If the student does not exist,
-- return null.
aFirstName Student.StdFirstName%type;
aLastName Student.StdLastName%type;

BEGIN

SELECT StdFirstName, StdLastName
 INTO aFirstName, aLastName
 FROM Student
 WHERE StdSSN = aStdSSN;

RETURN(aLastName || ', ' || aFirstName);

EXCEPTION
-- No_Data_Found is raised if the SELECT statement returns no data.
 WHEN No_Data_Found THEN
  RETURN(NULL);

 WHEN OTHERS THEN
   raise_application_error(-20001, 'Database error');

END;
/
-- Testing code
SET SERVEROUTPUT ON;
DECLARE
aStdName VARCHAR2(50);
BEGIN
-- This call should display a student name.
aStdName := fn_RetrieveStdName('901-23-4567');
IF aStdName IS NULL THEN
	dbms_output.put_line('Student not found');
ELSE
	dbms_output.put_line('Name is ' || aStdName);
END IF;

-- This call should not display a student name.
aStdName := fn_RetrieveStdName('905-23-4567');
IF aStdName IS NULL THEN
	dbms_output.put_line('Student not found');
ELSE
	dbms_output.put_line('Name is ' || aStdName);
END IF;
END;
/

11.16
CREATE OR REPLACE FUNCTION fn_ComputeWeightedGPA
(aStdSSN IN Student.StdSSN%type, aYear IN Offering.OffYear%Type) 
   RETURN NUMBER IS
-- Computes the weighted GPA given a student SSN and year.
-- Weighted GPA is the sum of units times the grade
-- divided by the total units.
-- If the student does not exist, return null.
WeightedGPA NUMBER;

BEGIN
SELECT SUM(enrgrade*CrsUnits)/SUM(CrsUnits)
 INTO WeightedGPA
 FROM Student, Registration, Enrollment, Offering, Course
 WHERE Student.StdSSN = aStdSSN
   AND Offering.OffYear = aYear
   AND Student.StdSSN = Registration.StdSSN
   AND Registration.RegNo = Enrollment.RegNo
   AND Enrollment.OfferNo = Offering.OfferNo
   AND Offering.CourseNo = Course.CourseNo;

RETURN(WeightedGPA);

EXCEPTION
 WHEN No_Data_Found THEN
  RETURN(NULL);

 WHEN OTHERS THEN
   raise_application_error(-20001, 'Database error');

END;
/
-- Testing code
SET SERVEROUTPUT ON;
DECLARE
aGPA DECIMAL(3,2);
BEGIN
-- This call should display a weighted GPA.
aGPA := fn_ComputeWeightedGPA('901-23-4567', 2006);
IF aGPA IS NULL THEN
	dbms_output.put_line('Student or enrollments not found');
ELSE
	dbms_output.put_line('Weighted GPA is ' || to_char(aGPA));
END IF;

-- This call should not display a weighted GPA.
aGPA := fn_ComputeWeightedGPA('905-23-4567', 2006);
IF aGPA IS NULL THEN
	dbms_output.put_line('Student or enrollments not found');
ELSE
	dbms_output.put_line('Weighted GPA is ' || to_char(aGPA));
END IF;
END;
/
-- Use the function in a query
SELECT StdSSN, StdFirstName, StdLastName,
       fn_ComputeWeightedGPA(StdSSN, 2006) AS WeightedGPA
FROM Student;

11.17
CREATE OR REPLACE FUNCTION fn_DetermineRank
(aStdSSN IN Student.StdSSN%TYPE, anOfferNo IN Offering.OfferNo%TYPE) 
   RETURN INTEGER IS
-- Determines the class rank for a given a student SSN and OfferNo.
-- Uses an implicit cursor.
-- If the student or offering do not exist, return 0.
TmpRank INTEGER :=0;
PrevEnrGrade Enrollment.EnrGrade%TYPE := 9.9;
FOUND BOOLEAN := FALSE;

BEGIN
-- Loop through implicit cursor
FOR EnrollRec IN 
 ( SELECT Student.StdSSN, EnrGrade
    FROM Student, Registration, Enrollment
    WHERE Enrollment.OfferNo = anOfferNo
      AND Student.StdSSN = Registration.StdSSN
      AND Registration.RegNo = Enrollment.RegNo
    ORDER BY EnrGrade DESC ) LOOP

    IF EnrollRec.EnrGrade < PrevEnrGrade THEN
    -- Increment the class rank when the grade changes
       TmpRank := TmpRank + 1;
       PrevEnrGrade := EnrollRec.EnrGrade;
    END IF;
    IF EnrollRec.StdSSN = aStdSSN THEN
       Found := TRUE;
       EXIT;
    END IF;
END LOOP;

IF Found THEN
  RETURN(TmpRank);
ELSE
  RETURN(0);
END IF;

EXCEPTION
WHEN OTHERS THEN
   raise_application_error(-20001, 'Database error');

END;
/
-- Testing code
SET SERVEROUTPUT ON;
-- Execute query to see test data
SELECT Student.StdSSN, EnrGrade
 FROM Student, Registration, Enrollment
 WHERE Enrollment.OfferNo = 5679
   AND Student.StdSSN = Registration.StdSSN
   AND Registration.RegNo = Enrollment.RegNo
 ORDER BY EnrGrade DESC;

-- Test script
DECLARE
aRank INTEGER;
BEGIN
-- This call should return a rank of 6.
aRank := fn_DetermineRank('789-01-2345', 5679);
IF aRank > 0 THEN
	dbms_output.put_line('Rank is ' || to_char(aRank));
ELSE
	dbms_output.put_line('Student is not enrolled.');
END IF;

-- This call should return a rank of 0.
aRank := fn_DetermineRank('789-01-2005', 5679);
IF aRank > 0 THEN
	dbms_output.put_line('Rank is ' || to_char(aRank));
ELSE
	dbms_output.put_line('Student is not enrolled.');
END IF;
END;
/

11.18
CREATE OR REPLACE PROCEDURE pr_DetermineRank
(aStdSSN IN Student.StdSSN%TYPE, anOfferNo IN Offering.OfferNo%TYPE,
 OutRank OUT INTEGER, OutGrade OUT Enrollment.EnrGrade%TYPE ) IS
-- Determines the class rank and grade for a given a student SSN
-- and OfferNo using an explicit cursor.
-- If the student or offering do not exist, return 0.
TmpRank INTEGER :=0;
PrevEnrGrade Enrollment.EnrGrade%TYPE := 9.9;
Found BOOLEAN := FALSE;
TmpGrade Enrollment.EnrGrade%TYPE;
TmpStdSSN Student.StdSSN%TYPE;
-- Explicit cursor
CURSOR EnrollCursor (tmpOfferNo Offering.OfferNo%TYPE) IS
   SELECT Student.StdSSN, EnrGrade
    FROM Student, Registration, Enrollment
    WHERE Enrollment.OfferNo = anOfferNo
      AND Student.StdSSN = Registration.StdSSN
      AND Registration.RegNo = Enrollment.RegNo
    ORDER BY EnrGrade DESC;

BEGIN
-- Open and loop through explicit cursor
OPEN EnrollCursor(anOfferNo);
LOOP
    FETCH EnrollCursor INTO TmpStdSSN, TmpGrade;
    EXIT WHEN EnrollCursor%NotFound;
    IF TmpGrade < PrevEnrGrade THEN
    -- Increment the class rank when the grade changes
       TmpRank := TmpRank + 1;
       PrevEnrGrade := TmpGrade;
    END IF;
    IF TmpStdSSN = aStdSSN THEN
       Found := TRUE;
       EXIT;
    END IF;
END LOOP;

CLOSE EnrollCursor;
IF Found THEN
  OutRank := TmpRank;
  OutGrade := PrevEnrGrade;
ELSE
  OutRank := 0;
  OutGrade := 0;
END IF;

EXCEPTION
WHEN OTHERS THEN
  raise_application_error(-20001, 'Database error');
END;
/
-- Testing code
SET SERVEROUTPUT ON;
-- Execute query to see test data
SELECT Student.StdSSN, EnrGrade
 FROM Student, Registration, Enrollment
 WHERE Student.StdSSN = Registration.StdSSN
   AND Registration.RegNo = Enrollment.RegNo
   AND Enrollment.OfferNo = 5679
 ORDER BY EnrGrade DESC;

-- Test script
DECLARE
aRank INTEGER;
aGrade Enrollment.EnrGrade%TYPE;
BEGIN
-- This call should produce a rank of 6.
pr_DetermineRank('789-01-2345', 5679, aRank, aGrade);
IF aRank > 0 THEN
	dbms_output.put_line('Rank is ' || to_char(aRank) || '.');
	dbms_output.put_line('Grade is ' || to_char(aGrade) || '.');
ELSE
	dbms_output.put_line('Student is not enrolled.');
END IF;

-- This call should produce a rank of 0.
pr_DetermineRank('789-01-2005', 5679, aRank, aGrade);
IF aRank > 0 THEN
	dbms_output.put_line('Rank is ' || to_char(aRank) || '.');
	dbms_output.put_line('Grade is ' || to_char(aGrade) || '.');
ELSE
	dbms_output.put_line('Student is not enrolled.');
END IF;
END;
/

11.19
CREATE OR REPLACE PACKAGE pck_University IS
PROCEDURE pr_DetermineRank
 (aStdSSN IN Student.StdSSN%TYPE, anOfferNo IN Offering.OfferNo%TYPE,
   OutRank OUT INTEGER, OutGrade OUT Enrollment.EnrGrade%TYPE );
FUNCTION fn_ComputeWeightedGPA
(aStdSSN IN Student.StdSSN%type, aYear IN Offering.OffYear%Type) 
   RETURN NUMBER;
END pck_University;
/

11.20
CREATE OR REPLACE PACKAGE BODY pck_University IS
PROCEDURE pr_DetermineRank
 (aStdSSN IN Student.StdSSN%TYPE, anOfferNo IN Offering.OfferNo%TYPE,
   OutRank OUT INTEGER, OutGrade OUT Enrollment.EnrGrade%TYPE ) IS
-- Determines the class rank and grade for a given a student SSN
-- and OfferNo using an explicit cursor.
-- If the student or offering do not exist, return 0.
TmpRank INTEGER :=0;
PrevEnrGrade Enrollment.EnrGrade%TYPE := 9.9;
Found BOOLEAN := FALSE;
TmpGrade Enrollment.EnrGrade%TYPE;
TmpStdSSN Student.StdSSN%TYPE;
-- Explicit cursor
CURSOR EnrollCursor (tmpOfferNo Offering.OfferNo%TYPE) IS
   SELECT Student.StdSSN, EnrGrade
    FROM Student, Registration, Enrollment
    WHERE Enrollment.OfferNo = anOfferNo
      AND Student.StdSSN = Registration.StdSSN
      AND Registration.RegNo = Enrollment.RegNo
    ORDER BY EnrGrade DESC;

BEGIN
-- Open and loop through explicit cursor
OPEN EnrollCursor(anOfferNo);
LOOP
    FETCH EnrollCursor INTO TmpStdSSN, TmpGrade;
    EXIT WHEN EnrollCursor%NotFound;
    IF TmpGrade < PrevEnrGrade THEN
    -- Increment the class rank when the grade changes
       TmpRank := TmpRank + 1;
       PrevEnrGrade := TmpGrade;
    END IF;
    IF TmpStdSSN = aStdSSN THEN
       Found := TRUE;
       EXIT;
    END IF;
END LOOP;

CLOSE EnrollCursor;
IF Found THEN
  OutRank := TmpRank;
  OutGrade := PrevEnrGrade;
ELSE
  OutRank := 0;
  OutGrade := 0;
END IF;

EXCEPTION
WHEN OTHERS THEN
  raise_application_error(-20001, 'Database error');
END pr_DetermineRank;

FUNCTION fn_ComputeWeightedGPA
(aStdSSN IN Student.StdSSN%type, aYear IN Offering.OffYear%Type) 
   RETURN NUMBER IS
-- Computes the weighted GPA given a student SSN and year.
-- Weighted GPA is the sum of units times the grade
-- divided by the total units.
-- If the student does not exist, return null.
WeightedGPA NUMBER;

BEGIN

SELECT SUM(enrgrade*CrsUnits)/SUM(CrsUnits)
 INTO WeightedGPA
 FROM Student, Registration, Enrollment, Offering, Course
 WHERE Student.StdSSN = aStdSSN
   AND Offering.OffYear = aYear
   AND Student.StdSSN = Registration.StdSSN
   AND Registration.RegNo = Enrollment.RegNo
   AND Enrollment.OfferNo = Offering.OfferNo
   AND Offering.CourseNo = Course.CourseNo;

RETURN(WeightedGPA);

EXCEPTION
 WHEN no_data_found THEN
  RETURN(NULL);

 WHEN OTHERS THEN
   raise_application_error(-20001, 'Database error');

END fn_ComputeWeightedGPA;

END pck_University;
/

11.21
SET SERVEROUTPUT ON;
DECLARE
aRank INTEGER;
aGrade Enrollment.EnrGrade%TYPE;
aGPA NUMBER;
BEGIN
-- This call should produce a rank of 6.
pck_University.pr_DetermineRank('789-01-2345', 5679, aRank, aGrade);
IF aRank > 0 THEN
	dbms_output.put_line('Rank is ' || to_char(aRank) || '.');
	dbms_output.put_line('Grade is ' || to_char(aGrade) || '.');
ELSE
	dbms_output.put_line('Student is not enrolled.');
END IF;
-- This call should display a weighted GPA.
aGPA := pck_University.fn_ComputeWeightedGPA('901-23-4567', 2006);
IF aGPA IS NULL THEN
	dbms_output.put_line('Student or enrollments not found');
ELSE
	dbms_output.put_line('Weighted GPA is ' || to_char(aGPA));
END IF;
END;
/

11.22
CREATE OR REPLACE TRIGGER tr_Course_IA
AFTER INSERT
ON Course
FOR EACH ROW
BEGIN
    -- No references to OLD row because only NEW exists for INSERT
    dbms_output.put_line('Inserted Row');
    dbms_output.put_line('CourseNo: ' || :NEW.CourseNo);
    dbms_output.put_line('Course Description: ' || :NEW.CrsDesc);
    dbms_output.put_line('Course Units: ' || To_Char(:NEW.CrsUnits));
END;
/
-- Testing statements
SET SERVEROUTPUT ON;
INSERT INTO Course (CourseNo, CrsDesc, CrsUnits)
VALUES ('IS485','Advanced Database Management',4);

ROLLBACK;

11.23
CREATE OR REPLACE TRIGGER tr_Course_UA
AFTER UPDATE
ON Course
FOR EACH ROW
BEGIN
    dbms_output.put_line('New Row Values');
    dbms_output.put_line('CourseNo: ' || :NEW.CourseNo);
    dbms_output.put_line('Course Description: ' || :NEW.CrsDesc);
    dbms_output.put_line('Course Units: ' || To_Char(:NEW.CrsUnits));

    dbms_output.put_line('Old Row Values');
    dbms_output.put_line('CourseNo: ' || :OLD.CourseNo);
    dbms_output.put_line('Course Description: ' || :OLD.CrsDesc);
    dbms_output.put_line('Course Units: ' || To_Char(:OLD.CrsUnits));
END;
/
-- Testing statements
SET SERVEROUTPUT ON;
-- Add row so it can be updated
INSERT INTO Course (CourseNo, CrsDesc, CrsUnits)
VALUES ('IS485','Advanced Database Management',4);

UPDATE Course
 SET CrsUnits = 3
 WHERE CourseNo = 'IS485';

ROLLBACK;

11.24
CREATE OR REPLACE TRIGGER tr_Course_DA
AFTER DELETE
ON Course
FOR EACH ROW
BEGIN
    -- No references to NEW row because only OLD exists for DELETE
    dbms_output.put_line('Deleted Row');
    dbms_output.put_line('CourseNo: ' || :OLD.CourseNo);
    dbms_output.put_line('Course Description: ' || :OLD.CrsDesc);
    dbms_output.put_line('Course Units: ' || To_Char(:OLD.CrsUnits));
END;
/
-- Testing statements
SET SERVEROUTPUT ON;
-- Insert row so that it can be deleted
INSERT INTO Course (CourseNo, CrsDesc, CrsUnits)
VALUES ('IS485','Advanced Database Management',4);

DELETE FROM Course
WHERE CourseNo = 'IS485';

ROLLBACK;

11.25
CREATE OR REPLACE TRIGGER tr_Course_DIUA
AFTER INSERT OR UPDATE OR DELETE
ON Course
FOR EACH ROW
BEGIN
    dbms_output.put_line('Inserted Table');
    dbms_output.put_line('CourseNo: ' || :NEW.CourseNo);
    dbms_output.put_line('Course Description: ' || :NEW.CrsDesc);
    dbms_output.put_line('Course Units: ' || To_Char(:NEW.CrsUnits));

    dbms_output.put_line('Deleted Table');
    dbms_output.put_line('CourseNo: ' || :OLD.CourseNo);
    dbms_output.put_line('Course Description: ' || :OLD.CrsDesc);
    dbms_output.put_line('Course Units: ' || To_Char(:OLD.CrsUnits));
END;
/
-- Testing statements
SET SERVEROUTPUT ON;
INSERT INTO Course (CourseNo, CrsDesc, CrsUnits)
VALUES ('IS485','Advanced Database Management',4);

UPDATE Course
SET CrsUnits = 3
WHERE CourseNo = 'IS485';

DELETE FROM Course
WHERE CourseNo = 'IS485';

ROLLBACK;

11.26
CREATE OR REPLACE TRIGGER tr_Enrollment_IB
-- This trigger ensures that the number of enrolled
-- students is less than the offering limit.
BEFORE INSERT
ON Enrollment
FOR EACH ROW
DECLARE 
   anOffLimit Offering.OffLimit%TYPE;
   anOffNumEnrolled Offering.OffNumEnrolled%TYPE;
   -- user defined exception declaration
   NoSeats EXCEPTION;
   ExMessage VARCHAR(200);
BEGIN
   SELECT OffLimit, OffNumEnrolled
     INTO anOffLimit, anOffNumEnrolled
     FROM Offering
     WHERE Offering.OfferNo = :NEW.OfferNo;

   IF anOffLimit <= anOffNumEnrolled THEN
      RAISE NoSeats;
   END IF;
EXCEPTION
  WHEN NoSeats THEN
   -- error number between -20000 and -20999
     ExMessage := 'No seats remaining in offering ' || 
                   to_char(:NEW.OfferNo) || '.';
     ExMessage := ExMessage || 'Number enrolled: ' || 
              to_char(anOffNumEnrolled) || '. ';
     ExMessage := ExMessage || 'Offering limit: ' || 
              to_char(anOffLimit);
     Raise_Application_Error(-20001, ExMessage);
END;
/
-- Testing statements
SET SERVEROUTPUT ON;
-- See offering limit and number enrolled
SELECT OffLimit, OffNumEnrolled
 FROM Offering
 WHERE Offering.OfferNo = 5679;
-- Insert the last student
INSERT INTO Enrollment (RegNo, OfferNo, EnrGrade)
VALUES (1234,5679,0);

-- update the number of enrolled students
UPDATE Offering
 SET OffNumEnrolled = OffNumEnrolled + 1
 WHERE OfferNo = 5679;

-- See offering limit and number enrolled
SELECT OffLimit, OffNumEnrolled
 FROM Offering
 WHERE Offering.OfferNo = 5679;
-- Insert a student beyond the limit
INSERT INTO Enrollment (RegNo, OfferNo, EnrGrade)
VALUES (1236,5679,0);

ROLLBACK;

11.27
CREATE OR REPLACE TRIGGER tr_Enrollment_IA
-- This trigger updates the number of enrolled
-- students the related offering row.
AFTER INSERT
ON Enrollment
FOR EACH ROW
BEGIN
   UPDATE Offering
    SET OffNumEnrolled = OffNumEnrolled + 1
    WHERE OfferNo = :NEW.OfferNo;
EXCEPTION
  WHEN OTHERS THEN
    RAISE_Application_Error(-20001, 'Database error');
END;
/
-- Testing statements
SET SERVEROUTPUT ON;
-- See the offering limit and number enrolled
SELECT OffLimit, OffNumEnrolled
 FROM Offering
 WHERE Offering.OfferNo = 5679;
-- Insert the last student
INSERT INTO Enrollment (RegNo, OfferNo, EnrGrade)
VALUES (1234,5679,0);

-- See the offering limit and number enrolled
SELECT OffLimit, OffNumEnrolled
 FROM Offering
 WHERE Offering.OfferNo = 5679;

ROLLBACK;

11.28
-- Drop the previous trigger to avoid interactions
DROP TRIGGER tr_Enrollment_IB;
CREATE OR REPLACE TRIGGER tr_Enrollment_IUB
-- This trigger ensures that the number of enrolled
-- students is less than the offering limit.
BEFORE INSERT OR UPDATE OF OfferNo
ON Enrollment
FOR EACH ROW
DECLARE 
    anOffLimit Offering.OffLimit%TYPE;
    anOffNumEnrolled Offering.OffNumEnrolled%TYPE;
	NoSeats EXCEPTION;
	ExMessage VARCHAR(200);
BEGIN
   SELECT OffLimit, OffNumEnrolled
     INTO anOffLimit, anOffNumEnrolled
     FROM Offering
     WHERE Offering.OfferNo = :NEW.OfferNo;

   IF anOffLimit <= anOffNumEnrolled THEN
      RAISE NoSeats;
   END IF;
EXCEPTION
  WHEN NoSeats THEN
   -- error number between -20000 and -20999
     ExMessage := 'No seats remaining in offering ' || 
                   to_char(:NEW.OfferNo) || '.';
     ExMessage := ExMessage || 'Number enrolled: ' || 
              to_char(anOffNumEnrolled) || '. ';
     ExMessage := ExMessage || 'Offering limit: ' || 
              to_char(anOffLimit);
     raise_application_error(-20001, ExMessage);
END;

11.29
-- Drop the previous trigger to avoid interactions
DROP TRIGGER tr_Enrollment_IA;
CREATE OR REPLACE TRIGGER tr_Enrollment_DIUA
-- This trigger updates the number of enrolled
-- students the related offering row.
AFTER INSERT OR DELETE OR UPDATE of OfferNo
ON Enrollment
FOR EACH ROW
BEGIN
 -- Increment the number of enrolled students for insert, update
 IF INSERTING OR UPDATING THEN   
    UPDATE Offering
      SET OffNumEnrolled = OffNumEnrolled + 1
      WHERE OfferNo = :NEW.OfferNo;
 END IF;
-- Decrease the number of enrolled students for delete, update
IF UPDATING OR DELETING THEN
    UPDATE Offering
      SET OffNumEnrolled = OffNumEnrolled - 1
      WHERE OfferNo = :OLD.OfferNo;
END IF;

EXCEPTION
  WHEN OTHERS THEN
    raise_application_error(-20001, 'Database error');
END;

11.30
-- Test case 1
-- See the offering limit and number enrolled
SELECT OffLimit, OffNumEnrolled
 FROM Offering
 WHERE Offering.OfferNo = 5679;
-- Insert the last student
INSERT INTO Enrollment (RegNo, OfferNo, EnrGrade)
VALUES (1234,5679,0);
-- See the offering limit and the number enrolled
SELECT OffLimit, OffNumEnrolled
 FROM Offering
 WHERE Offering.OfferNo = 5679;

-- Test case 2
-- Insert a student beyond the limit: exception raised
INSERT INTO Enrollment (RegNo, OfferNo, EnrGrade)
VALUES (1236,5679,0);
-- Transfer a student to offer 5679: exception raised
UPDATE Enrollment
 SET OfferNo = 5679
 WHERE RegNo = 1234 AND OfferNo = 1234;

-- Test case 3
-- See the offering limit and the number enrolled before update
SELECT OffLimit, OffNumEnrolled
 FROM Offering
 WHERE Offering.OfferNo = 4444;
-- Update a student to a non full offering
UPDATE Enrollment
 SET OfferNo = 4444
 WHERE RegNo = 1234 AND OfferNo = 1234;
-- See the offering limit and the number enrolled after update
SELECT OffLimit, OffNumEnrolled
 FROM Offering
 WHERE Offering.OfferNo = 4444;

-- Test case 4
-- See the offering limit and the number enrolled before delete
SELECT OffLimit, OffNumEnrolled
 FROM Offering
 WHERE Offering.OfferNo = 1234;
-- Delete an enrollment
DELETE Enrollment
 WHERE OfferNo = 1234;
-- See the offering limit and the number enrolled
SELECT OffLimit, OffNumEnrolled
 FROM Offering
 WHERE Offering.OfferNo = 1234;

-- Erase all changes
ROLLBACK;

11.31
CREATE OR REPLACE TRIGGER tr_FacultySalary_UB
-- This trigger ensures that a salary increase does not exceed
-- 10%.
BEFORE UPDATE OF FacSalary
ON Faculty
FOR EACH ROW
WHEN (NEW.FacSalary > 1.1 * OLD.FacSalary)
DECLARE 
	SalaryIncreaseTooHigh EXCEPTION;
	ExMessage VARCHAR(200);
BEGIN
    RAISE SalaryIncreaseTooHigh;
EXCEPTION
  WHEN SalaryIncreaseTooHigh THEN
   -- error number between -20000 and -20999
     ExMessage := 'Salary increase exceeds 10%. '; 
     ExMessage := ExMessage || 'Current salary: ' || 
              to_char(:OLD.FacSalary) || '. ';
     ExMessage := ExMessage || 'New salary: ' || 
              to_char(:NEW.FacSalary) || '.';
     Raise_Application_Error(-20001, ExMessage);
END;
/
SET SERVEROUTPUT ON;
-- Test case 1: salary increase of 5%
UPDATE Faculty
 SET FacSalary = FacSalary * 1.05
 WHERE FacSSN = '543-21-0987';
SELECT FacSalary FROM Faculty WHERE FacSSN = '543-21-0987';
-- Test case 2: salary increase of 20% should generate an exception.
UPDATE Faculty
 SET FacSalary = FacSalary * 1.20
 WHERE FacSSN = '543-21-0987';
ROLLBACK;

11.32
CREATE OR REPLACE TRIGGER tr_FacultyName_IUB
-- This trigger changes the case of FacFirstName and FacLastName.
BEFORE INSERT OR UPDATE OF FacFirstName, FacLastName
ON Faculty
FOR EACH ROW
BEGIN
   :NEW.FacFirstName := Upper(:NEW.FacFirstName);
   :NEW.FacLastName := Upper(:NEW.FacLastName);
END;
/
-- Testing statements
UPDATE Faculty
 SET FacFirstName = 'Joe', FacLastName = 'Smith'
 WHERE FacSSN = '543-21-0987';
-- Display the changed faculty name.
SELECT FacFirstName, FacLastName
 FROM Faculty
 WHERE FacSSN = '543-21-0987';
ROLLBACK;
